#******************************************************************************************************
# Python code for predicting thermochemical properties of alkanes using linear regression
#******************************************************************************************************

#******************************************************************************************************
'''
MIT License

Copyright (c) 2024 Shrinjay Sharma, Josh Sleijfer, Jeroen Op de Beek, Stach van der Zeeuw, Daniil Zorzos, Silvia Lasala, Marcello S. Rigutto, Erik Zuidema, Umang Agarwal, Richard Baur, Sofia Calero, David Dubbeldam, and Thijs J.H. Vlugt

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''
#******************************************************************************************************

#******************************************************************************************************
# This code reads data from the Supplementary Information SI1.xlsx. SI1.xlsx must be in the same folder as the code.

# The following libraries must be installed beforehand to run the code:
# pandas, numpy, math, matplotlib, sklearn, typing, itertools, more_itertools, rdkit

# names of thermochemical properties:
#"DHf0"        - Enthalpy of formation / [kJ/mol]
#"DGf0"        - Gibbs free energy of formation / [kJ/mol]
#"(G0-H0(0K))" - Gibbs free energy content / [kJ/mol]
#"H0-H0(0K)"   - Enthalpy content / [kJ/mol]

# temperature
# It can range from 0 to 1000 K : 0, 200, 273.15, 298.15, 300, 400, 500, 600, 700, 800, 900, 1000

# Guidelines to write abbreviations for alkane isomers
# It follows IUPAC conventions in numbering the positions of the branches 

# type of branch                           - symbol used
# mthyl               			   - m
# ethyl               			   - e
# propyl              			   - p
# isopropyl           			   - ip
# butyl               			   - b
# secondary butyl (1-methylpropyl)         - sb
# tertiary butyl (1,1-dimethylethyl)       - tb
# isobutyl (2-methylpropyl)                - ib
# pentyl                                   - pen
# tertiary pentyl (1,1-dimethylpropyl)     - tpen  
# neopentyl (2,2-dimethylpropyl)           - npen  
# isopentyl (3-methylbutyl)                - ipen  
# secondary pentyl (1-methylbutyl)         - spen  
# 3 pentyl (1-ethylpropyl)                 - iiipen 
# secondary isopentyl (1,2-dimethylpropyl) - sipen
# active pentyl (2-methylbutyl)            - actpen

# Example: 4-(1,1-Dimethylethyl)-5-ethyl-2-methylheptane: "4-tb-5-e-2-m-C7"


# Examples 
# To run an example, copy and paste this snippet to the end of the code
'''
# Identify which second order groups are present in a molecule
print(featureNames) #list of second order groups
print(get_second_order_feature(iupac_smiles("3-e-2,2,4,4-m-3-tb-C5"))) #gives the occurrences of the second order groups

# Computing thermochemical properties using linear regression


temperatures = [0, 200, 273.15, 298.15, 300, 400, 500, 600, 700, 800, 900, 1000]

model_dict = mylinearregression(featureNames,
                   DataToUse=11,
                   filterUnknown=True,
                   temperature_range = [temperatures[0], temperatures[-1]]
                   )
# property obtained from the training dataset (In this case Enthalpy of formation DHf0 for 3,3-m-C8 at 400 K is obtained from the Scott's tables)
#print("isomer-abbreviation Training dataset:", get_scott_data(concatenated_df, "isomer abbreviation", temperature, "thermochemical property")) 
print("3,3-m-C8 Training dataset:", get_scott_data(concatenated_df, "3,3-m-C8", 400, "DHf0")) 

# property predicted using the LR model with second order group contributions (In this case Enthalpy of formation DHf0 for 3,3-m-C8 at 400 K is computed using the LR model)
#print("isomer abbreviation Linear regression:", get_fit_value(model_dict, "thermochemical property", temperature, temperatures, "isomer abbreviation")) 
print("3,3-m-C8 Linear regression:", get_fit_value(model_dict, "DHf0", 400, temperatures, "3,3-m-C8")) 
'''
#*****************************************************************************************************


#******************************************************************************************************
# Import libraries
#******************************************************************************************************

import pandas as pd
import numpy as np
import math

import matplotlib.pyplot as plt

from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error


from sklearn.model_selection import train_test_split
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import StandardScaler

from math import comb
from typing import Dict,List,Tuple
import itertools
import more_itertools

from rdkit import Chem

#*****************************************************************************************************
# Converting IUPAC abbreviations of alkanes to SMILES strings:
# Valid till isomers with pentyl groups as branches
#*****************************************************************************************************

def iupac_smiles(molecule):
    s = molecule.split("-")
    s1 = s[len(s)-1].split("C")
    
    Nchain = int(s1[len(s1)-1]) # length of the molecule
    #print('Nchain: ', Nchain)
    b_name = [] # initializing the lit of possible branch names
    
    sd_m, sd_e, sd_p, sd_b, sd_ip, sd_ib = [], [], [], [], [], []
    sd_sb, sd_tb, sd_pen, sd_tpen, sd_npen = [], [], [], [], []
    sd_ipen, sd_spen, sd_iiipen, sd_sipen, sd_actpen = [], [], [], [], []
    
    for i in range(1,len(s),2):
        sdummy = s[i-1]
        comma_find = sdummy.find(",") # finding if there is a comma
        
        if sdummy[comma_find] == ",":
            if s[i] == "m":
                sd_m = s[i-1].split(",")
            elif s[i] == "e":
                sd_e = s[i-1].split(",")
            elif s[i] == "p":
                sd_p = s[i-1].split(",")
            elif s[i] == "b":
                sd_b = s[i-1].split(",")
            elif s[i] == "ip":
                sd_ip = s[i-1].split(",")
            elif s[i] == "ib":
                sd_ib = s[i-1].split(",")
            elif s[i] == "sb":
                sd_sb = s[i-1].split(",")
            elif s[i] == "tb":
                sd_tb = s[i-1].split(",")
            elif s[i] == "pen":
                sd_pen = s[i-1].split(",")
            elif s[i] == "tpen":
                sd_tpen = s[i-1].split(",")
            elif s[i] == "npen":
                sd_npen = s[i-1].split(",")
            elif s[i] == "ipen":
                sd_ipen = s[i-1].split(",")
            elif s[i] == "spen":
                sd_spen = s[i-1].split(",")
            elif s[i] == "iiipen":
                sd_iiipen = s[i-1].split(",")
            elif s[i] == "sipen":
                sd_sipen = s[i-1].split(",")
            elif s[i] == "actpen":
                sd_actpen = s[i-1].split(",")
        else:
            if s[i] == "m":
                sd_m.append(s[i-1])
            elif s[i] == "e":
                sd_e.append(s[i-1])
            elif s[i] == "p":
                sd_p.append(s[i-1])
            elif s[i] == "b":
                sd_b.append(s[i-1])
            elif s[i] == "ip":
                sd_ip.append(s[i-1])
            elif s[i] == "ib":
                sd_ib.append(s[i-1])
            elif s[i] == "sb":
                sd_sb.append(s[i-1])
            elif s[i] == "tb":
                sd_tb.append(s[i-1])
            elif s[i] == "pen":
                sd_pen = s[i-1]
            elif s[i] == "tpen":
                sd_tpen.append(s[i-1])
            elif s[i] == "npen":
                sd_npen.append(s[i-1])
            elif s[i] == "ipen":
                sd_ipen.append(s[i-1])
            elif s[i] == "spen":
                sd_spen.append(s[i-1])
            elif s[i] == "iiipen":
                sd_iiipen.append(s[i-1])
            elif s[i] == "sipen":
                sd_sipen.append(s[i-1])
            elif s[i] == "actpen":
                sd_actpen.append(s[i-1])
                
        b_name.append(s[i])
          
    bs_m, bs_e, bs_p, bs_b, bs_ip, bs_ib = "(C)", "(CC)", "(CCC)", "(CCCC)", "(C(C)C)", "(CC(C)C)"
    bs_sb, bs_tb = "(C(C)CC)", "(C(C)(C)C)"
    bs_pen, bs_tpen, bs_npen = "(CCCCC)", "(C(C)(C)CC)", "(CC(C)(C)C)"
    bs_ipen, bs_spen, bs_iiipen, bs_sipen, bs_actpen = "(CCC(C)C)", "(C(C)CCC)", "(C(CC)CC)", "(C(C)C(C)C)", "(CC(C)CC)"
    #bs = ["(C)", "(CC)", "(CCC)", "(CCCC)", "(C(C)C)", "(C(C)CC)"]
    if Nchain == 1:
        sm_st = "C"
    else:
        sm_st = "CC"
        for k in range(1,Nchain-1):
            
            for bm in b_name:
                if bm=="m":
                    for j in range(0,len(sd_m)):
                        if sd_m[j]==str(k+1):
                            sm_st += bs_m
                        else:
                            sm_st = sm_st
                elif bm=="e":
                    
                    for j in range(0,len(sd_e)):
                        if sd_e[j]==str(k+1):
                            sm_st += bs_e
                        else:
                            sm_st = sm_st
                elif bm=="p":
                    
                    for j in range(0,len(sd_p)):
                        if sd_p[j]==str(k+1):
                            sm_st += bs_p
                        else:
                            sm_st = sm_st
                elif bm=="b":
                    
                    for j in range(0,len(sd_b)):
                        if sd_b[j]==str(k+1):
                            sm_st += bs_b
                        else:
                            sm_st = sm_st
                elif bm=="ip":
                    
                    for j in range(0,len(sd_ip)):
                        if sd_ip[j]==str(k+1):
                            sm_st += bs_ip
                        else:
                            sm_st = sm_st
                elif bm=="ib":
                    
                    for j in range(0,len(sd_ib)):
                        if sd_ib[j]==str(k+1):
                            sm_st += bs_ib
                        else:
                            sm_st = sm_st
                elif bm=="sb":
                    
                    for j in range(0,len(sd_sb)):
                        if sd_sb[j]==str(k+1):
                            sm_st += bs_sb
                        else:
                            sm_st = sm_st
                elif bm=="tb":
                    
                    for j in range(0,len(sd_tb)):
                        if sd_tb[j]==str(k+1):
                            sm_st += bs_tb
                        else:
                            sm_st = sm_st
                elif bm=="pen":
                    
                    for j in range(0,len(sd_pen)):
                        if sd_pen[j]==str(k+1):
                            sm_st += bs_pen
                        else:
                            sm_st = sm_st
                elif bm=="tpen":
                    
                    for j in range(0,len(sd_tpen)):
                        if sd_tpen[j]==str(k+1):
                            sm_st += bs_tpen
                        else:
                            sm_st = sm_st
                elif bm=="npen":
                    
                    for j in range(0,len(sd_npen)):
                        if sd_npen[j]==str(k+1):
                            sm_st += bs_npen
                        else:
                            sm_st = sm_st
                elif bm=="ipen":
                    
                    for j in range(0,len(sd_ipen)):
                        if sd_ipen[j]==str(k+1):
                            sm_st += bs_ipen
                        else:
                            sm_st = sm_st
                elif bm=="spen":
                    
                    for j in range(0,len(sd_spen)):
                        if sd_spen[j]==str(k+1):
                            sm_st += bs_spen
                        else:
                            sm_st = sm_st
                elif bm=="iiipen":
                    
                    for j in range(0,len(sd_iiipen)):
                        if sd_iiipen[j]==str(k+1):
                            sm_st += bs_iiipen
                        else:
                            sm_st = sm_st
                elif bm=="sipen":
                    
                    for j in range(0,len(sd_sipen)):
                        if sd_sipen[j]==str(k+1):
                            sm_st += bs_sipen
                        else:
                            sm_st = sm_st
                elif bm=="actpen":
                    
                    for j in range(0,len(sd_actpen)):
                        if sd_actpen[j]==str(k+1):
                            sm_st += bs_actpen
                        else:
                            sm_st = sm_st
    
            sm_st += "C"
    return sm_st

#*****************************************************************************************************
# Reading and extracting tables of thermochemical properties from Excel workbook
#*****************************************************************************************************

# Read Scott's tables

def getAll_tables():
    #Reads the Thermodynamic data of alkanes from C1 to C10 from the digitalized copy of Scott's table
    #thermo_file = "Extended_data_set_Alkanes.xlsx" #variable storing the name of the excel file 
    thermo_file = "SI1.xlsx" #variable storing the name of the excel file 
    excel_file = pd.ExcelFile(thermo_file)

    # Initialize a dictionary to store tables from all sheets
    all_tables = {}

    table_rows = 19 #no of rows for each molecule (including the two rows of headings)
    table_cols = 5 #no of columns for each molecule (included DGf0 (kJ/mol))

    # Loop through each sheet
    for sheet_name in excel_file.sheet_names: #sheet_name: name of the sheet. It will go through every sheet
        sheet_df = excel_file.parse(sheet_name) #reading a particular sheet in the excel file

        # Find the location of the "T" cell to identify the starting point of each table
        for row in range(sheet_df.shape[0]): # loop over all rows in each excel sheet
            for col in range(sheet_df.shape[1]): # loop over all columns in each excel sheet
                if sheet_df.iat[row, col] == 'T': #search for a cells with with the word T -- Temperature
                    # Extract the table within the fixed size boundaries
                    # extract data from cells: row of the sheet + 
                    table_data = sheet_df.iloc[row:row + table_rows, col:col + table_cols] #a table for each component
                    table_data.columns = table_data.iloc[0] #assigning the first row (names of the columns) as headers

                    # Remove the first two rows (headers and units) 
                    table_data = table_data.iloc[2:]

                    # Store the extracted table in a dictionary
                    table_name = f"{sheet_name} {sheet_df.iat[row - 1, col]}"
                    all_tables[table_name] = table_data                    
    return all_tables
    
all_tables = getAll_tables()
new_tables = {} #dictionary for all molecules
for old_key, table_data in all_tables.items():
    # Split the old key by space and take the last part as the new key
    # The original names of the tables were sheet_name+molecule name
    # Remove the sheet name and just keep the molecule name
    myname = old_key.split()[-1]
    new_key = iupac_smiles(myname) #the new key is a smiles string
    table_data.columns = [str(col) for col in table_data.columns]
    new_tables[(myname,new_key)] = table_data
    
all_tables = new_tables

# Making everything into one table:, with an extra column associated with the molecule:
concatenated_df1 = pd.concat((df.assign(Molecule=table_name[1]) for table_name, \
                             df in all_tables.items()), ignore_index=True)

# Finding duplicates
#duplicates = concatenated_df1.index[concatenated_df1.index.duplicated()]
#print("Duplicate index values:", duplicates)

column_mol = concatenated_df1['Molecule']
column_mol_T = concatenated_df1['T']
column_mol1 = [str(element) for element in column_mol]


#*****************************************************************************************************
# Adding 2nd order groups to the dataframe:
#*****************************************************************************************************

types = 4 # CH3, CH2, CH, C
ans=0 #initializing the number of second order groups

def numHs(atom: Chem.Atom) -> int:
    return 4-len(atom.GetNeighbors())

secondOrderGroups: Dict[Tuple,int] = {}
secondOrderNames = [] #array initialization for 2nd order group names


for i in range(5): # 5 --> CH0, CH1, CH2, CH3, CH4
    for starsandstripes in more_itertools.distinct_permutations('*'*(4-i) + '|'*3):
        at = 0
        cur = [i]
        #myname is either C, CH1, CH2, CH3, CH4
        myname = "C"
        if i!=0: #i is the number of H atoms attached to C atom
            myname = f"CH{i}" #f --> format string
        for c in starsandstripes:
            if c=='*':
                cur.append(at)
            else:
                at+=1
        for at in cur[1:]:
            if at==0:
                myname+="(C)" # eg. C(C)(C)(C)(C)
            else:
                myname+=f"(CH{at})"

        secondOrderNames.append(myname)
        
        sz = len(secondOrderGroups) #length of the second order groups
        secondOrderGroups[tuple(cur)] = sz


for i in range(0,types+1):
    cur = comb(types-1+i,i)
    # print("with CH",4-i, ": ", cur,sep="")
    ans+=cur
    
N = len(secondOrderGroups) #length of the second order groups


def getTypesAll(mol: Chem.Mol) -> np.ndarray:  #povides the occurences of each second order groups for a molecule
    cnts = [0]*len(secondOrderGroups)
    for atom in mol.GetAtoms():
        mytype = tuple([numHs(atom)] + sorted(numHs(i) for i in atom.GetNeighbors()))
        cnts[secondOrderGroups[mytype]]+=1
    return np.array(cnts)


def get_second_order_feature(isomer): 
    featureNames = ['CH4','CH3(C)','CH3(CH)','CH3(CH2)','CH3(CH3)',
                    'CH2(C)(C)','CH2(C)(CH)','CH2(C)(CH2)','CH2(C)(CH3)',
                    'CH2(CH)(CH)','CH2(CH)(CH2)','CH2(CH)(CH3)','CH2(CH2)(CH2)',
                    'CH2(CH2)(CH3)','CH2(CH3)(CH3)','CH(C)(C)(CH3)','CH(C)(CH)(CH2)',
                    'CH(C)(CH)(CH3)','CH(C)(CH2)(CH2)','CH(C)(CH2)(CH3)','CH(C)(CH3)(CH3)',
                    'CH(CH)(CH)(CH)','CH(CH)(CH)(CH2)','CH(CH)(CH)(CH3)','CH(CH)(CH2)(CH2)',
                    'CH(CH)(CH2)(CH3)','CH(CH)(CH3)(CH3)','CH(CH2)(CH2)(CH2)','CH(CH2)(CH2)(CH3)',
                    'CH(CH2)(CH3)(CH3)','CH(CH3)(CH3)(CH3)','C(C)(CH)(CH3)(CH3)','C(C)(CH2)(CH2)(CH3)',
                    'C(C)(CH2)(CH3)(CH3)','C(C)(CH3)(CH3)(CH3)','C(CH)(CH)(CH2)(CH3)',
                    'C(CH)(CH)(CH3)(CH3)','C(CH)(CH2)(CH2)(CH2)','C(CH)(CH2)(CH2)(CH3)',
                    'C(CH)(CH2)(CH3)(CH3)','C(CH)(CH3)(CH3)(CH3)','C(CH2)(CH2)(CH2)(CH2)',
                    'C(CH2)(CH2)(CH2)(CH3)','C(CH2)(CH2)(CH3)(CH3)','C(CH2)(CH3)(CH3)(CH3)','C(CH3)(CH3)(CH3)(CH3)']

    # second order groups not present in the training dataset are approximated with another similar second order group which is present in the isomers of the training dataset

    sog_abs_1 = ['C(C)(C)(C)(C)','C(C)(C)(C)(CH1)','C(C)(C)(C)(CH2)','C(C)(C)(C)(CH3)',
                 'C(C)(C)(CH1)(CH1)','C(C)(C)(CH1)(CH2)','C(C)(C)(CH1)(CH3)','C(C)(C)(CH2)(CH2)',
                 'C(C)(C)(CH2)(CH3)','C(C)(C)(CH3)(CH3)','C(C)(CH1)(CH1)(CH1)','C(C)(CH1)(CH1)(CH2)',
                 'C(C)(CH1)(CH1)(CH3)','C(C)(CH1)(CH2)(CH2)','C(C)(CH1)(CH2)(CH3)'] # will be replaced by C(C)(CH)(CH3)(CH3)
    sog_abs_2 = ['C(C)(CH2)(CH2)(CH2)'] # will be replaced by C(C)(CH2)(CH2)(CH3)
    sog_abs_3 = ['C(CH1)(CH1)(CH1)(CH1)','C(CH1)(CH1)(CH1)(CH2)','C(CH1)(CH1)(CH1)(CH3)',
                 'C(CH1)(CH1)(CH2)(CH2)'] # will be replaced by C(CH)(CH)(CH2)(CH3)
    sog_abs_4 = ['CH1(C)(C)(C)','CH1(C)(C)(CH1)','CH1(C)(C)(CH2)'] # will be replaced by CH(C)(C)(CH3)
    sog_abs_5 = ['CH1(C)(CH1)(CH1)'] # will be replaced by CH(C)(CH)(CH2)
    
    sog1, sog2, sog3, sog4, sog5 = [], [], [], [], []
    featureNames1 = []
    # Making arrays of characters #[[ ]]
    for compound in featureNames:
        # Remove parentheses and split the compound into elements
        elements = compound.replace('(', ',').replace(')', '').split(',')
        # Append the list of elements to the result
        featureNames1.append(elements)
    
    for compound in sog_abs_1:
        # Remove parentheses and split the compound into elements
        elements = compound.replace('(', ',').replace(')', '').split(',')
        # Append the list of elements to the result
        sog1.append(elements)
    
    for compound in sog_abs_2:
        # Remove parentheses and split the compound into elements
        elements = compound.replace('(', ',').replace(')', '').split(',')
        # Append the list of elements to the result
        sog2.append(elements)
    
    for compound in sog_abs_3:
        # Remove parentheses and split the compound into elements
        elements = compound.replace('(', ',').replace(')', '').split(',')
        # Append the list of elements to the result
        sog3.append(elements)
    
    for compound in sog_abs_4:
        # Remove parentheses and split the compound into elements
        elements = compound.replace('(', ',').replace(')', '').split(',')
        # Append the list of elements to the result
        sog4.append(elements)
    
    for compound in sog_abs_5:
        # Remove parentheses and split the compound into elements
        elements = compound.replace('(', ',').replace(')', '').split(',')
        # Append the list of elements to the result
        sog5.append(elements)
            
    #****************************************************************************
    
    smiles_str_list = [char for char in isomer if char not in ('(', ')')]
    conv_smiles_list = []
    
    isomer = Chem.MolFromSmiles(isomer)
    
    for i in range(0,len(smiles_str_list)):
        atom = isomer.GetAtomWithIdx(i)
        # Get the neighbors
        neighbors = list(atom.GetNeighbors())
    
        nn=[] #neighbors
        nn_idx=[] #position number or index of neighbors
        if len(neighbors) == 0:
            conv_smiles_list.append("CH4")
        if len(neighbors) == 1:
            conv_smiles_list.append("CH3")
        elif len(neighbors) == 2:
            conv_smiles_list.append("CH2")
        elif len(neighbors) == 3:
            conv_smiles_list.append("CH")
        elif len(neighbors) == 4:
            conv_smiles_list.append("C")
    #*******************************************************************************
    
    count_list = np.zeros(len(featureNames),dtype=int)
    #count_list = {}
    
    for i in range(0,len(smiles_str_list)):
        atom = isomer.GetAtomWithIdx(i)
        # Get the neighbors
        neighbors = list(atom.GetNeighbors())
    
        nn=[] #neighbors
        nn_idx=[] #position number or index of neighbors
            
        for j in range(0,len(neighbors)):
            nn.append((neighbors[j]).GetSymbol())
            nn_idx.append((neighbors[j]).GetIdx())
            
        dummy_str = conv_smiles_list[i]
        
        dummy_str_list = []
        for k in nn_idx:
            dummy_str += '(' + conv_smiles_list[k] + ')'

    # Remove parentheses and split the compound into elements
        elements = dummy_str.replace('(', ',').replace(')', '').split(',')
        # Append the list of elements to the result
        dummy_str_list.append(elements)   #[ ]
    
        # Generate all permutations of the elements in dummy_str_list
        perm_dummy_str_list = set(itertools.permutations(dummy_str_list[0]))
        # Check if any permutation of dummy_str_list matches with any nested list in B
        
        matches = any(tuple(b) in perm_dummy_str_list for b in sog1)
        if matches:
            dummy_str_list = [['C','C','CH','CH3','CH3']]
            
        matches = any(tuple(b) in perm_dummy_str_list for b in sog2)    
        if matches:
            dummy_str_list = [['C','C','CH2','CH2','CH3']]
            
        matches = any(tuple(b) in perm_dummy_str_list for b in sog3)
        if matches:
            dummy_str_list = [['C','CH','CH','CH2','CH3']]
            
        matches = any(tuple(b) in perm_dummy_str_list for b in sog4)
        if matches:
            dummy_str_list = [['CH','C','C','CH3']]
            
        matches = any(tuple(b) in perm_dummy_str_list for b in sog5)
        if matches:
            dummy_str_list = [['CH','C','CH','CH2']]
        
        # Generate all permutations of the elements in dummy_str_list
        perm_dummy_str_list = set(itertools.permutations(dummy_str_list[0]))
    
        # Loop through elements in featureNames1
        for j in range(len(featureNames1)):
            # Convert the element to a tuple (since A contains tuples)
            element_tuple = tuple(featureNames1[j])
            
            # Check if the element is in the dictionary
            if element_tuple in perm_dummy_str_list:
                count_list[j] += 1
    count_list = list(count_list)
    return count_list
    
featureNames = ['CH4','CH3(C)','CH3(CH)','CH3(CH2)','CH3(CH3)',
                'CH2(C)(C)','CH2(C)(CH)','CH2(C)(CH2)','CH2(C)(CH3)',
                'CH2(CH)(CH)','CH2(CH)(CH2)','CH2(CH)(CH3)','CH2(CH2)(CH2)',
                'CH2(CH2)(CH3)','CH2(CH3)(CH3)','CH(C)(C)(CH3)','CH(C)(CH)(CH2)',
                'CH(C)(CH)(CH3)','CH(C)(CH2)(CH2)','CH(C)(CH2)(CH3)','CH(C)(CH3)(CH3)',
                'CH(CH)(CH)(CH)','CH(CH)(CH)(CH2)','CH(CH)(CH)(CH3)','CH(CH)(CH2)(CH2)',
                'CH(CH)(CH2)(CH3)','CH(CH)(CH3)(CH3)','CH(CH2)(CH2)(CH2)','CH(CH2)(CH2)(CH3)',
                'CH(CH2)(CH3)(CH3)','CH(CH3)(CH3)(CH3)','C(C)(CH)(CH3)(CH3)','C(C)(CH2)(CH2)(CH3)',
                'C(C)(CH2)(CH3)(CH3)','C(C)(CH3)(CH3)(CH3)','C(CH)(CH)(CH2)(CH3)',
                'C(CH)(CH)(CH3)(CH3)','C(CH)(CH2)(CH2)(CH2)','C(CH)(CH2)(CH2)(CH3)',
                'C(CH)(CH2)(CH3)(CH3)','C(CH)(CH3)(CH3)(CH3)','C(CH2)(CH2)(CH2)(CH2)',
                'C(CH2)(CH2)(CH2)(CH3)','C(CH2)(CH2)(CH3)(CH3)','C(CH2)(CH3)(CH3)(CH3)','C(CH3)(CH3)(CH3)(CH3)']



df_mol = concatenated_df1['Molecule'].tolist()

mol_count = [get_second_order_feature(df_mol[i]) for i in range(len(df_mol))]

df_mol_count = pd.DataFrame({col_name: [mol_count[i][j] for i in range(len(df_mol))] for j, col_name in enumerate(featureNames)})

concatenated_df = pd.concat([concatenated_df1, df_mol_count], axis=1)

#Removing duplicate columns
concatenated_df = concatenated_df.loc[:,~concatenated_df.columns.duplicated()]

#concatenated_df = concatenated_df1.merge(df_mol_count)
#removing the row 400K for 3mC8 because (H0-H0(0K)) is wrong in Scott's table 
concatenated_df = concatenated_df[(concatenated_df["T"]!=400.) | (concatenated_df["Molecule"]!=iupac_smiles("3-m-C8"))]

#*****************************************************************************************************
# Adding a column for the total count of carbon atoms present in an isomer
#*****************************************************************************************************

# Old features:
concatenated_df['Total Carbons'] = concatenated_df['Molecule'].str.count('C')

outputs = ['DHf0','DGf0','(G0-H0(0K))','H0-H0(0K)']


#*****************************************************************************************************
# Linear regression
#*****************************************************************************************************

import random

random.seed(69)


#glob_lgcoefs = None


def mylinearregression(myinputs,
                       filterUnknown=True,
                       DataToUse=11, 
                       temperature_range = [-1, 100000]
                       ) -> Dict:
    badMolecules = set()

    all_models = dict() # key: (value, temperature)

    for output_column in outputs:
        match output_column: # Making nicer shorter keys for the model dictionary:
            case "DHf0":
                value_name = "DHf0"
            case "DGf0":
                value_name = "DGf0"
            case "(G0-H0(0K))":
                value_name = "(G0-H0(0K))"
            case "H0-H0(0K)":
                value_name = "H0-H0(0K)"

        #lgcoefs = []
        lgtemp = []

        for temperature_value, group in concatenated_df.groupby('T'):
            if temperature_value < temperature_range[0] or temperature_value > temperature_range[1]:  #or temperature_value > 700:
                continue

            if filterUnknown:
                actual = group[group["Total Carbons"] == group[myinputs].sum(axis=1)]
                nope = group[group["Total Carbons"] != group[myinputs].sum(axis=1)]
                group = actual
            input_data = group[myinputs + ['Molecule']]

            output_data = group[outputs]

            
            X_train = input_data[group["Total Carbons"] < DataToUse].drop(
                    columns='Molecule')
           
            #***************************************************
                
            y_train = output_data[group["Total Carbons"] < DataToUse]

            #***************************************************
            
            # Create and fit the linear regression model
            model = LinearRegression()
            model.fit(X_train.values, y_train[output_column].values)
            #***************************************************
            #lgcoefs.append(model.coef_)
            #predictions = model.predict(X_test)
            #***************************************************
            #all_predictions = model.predict(input_data.drop(columns="Molecule"))
            all_models[(value_name, temperature_value)] = model
              
            lgtemp.append(temperature_value)
            
            
    return all_models


def find_two_closest(value: int | float, lst: list): # Assumes distinct list elements.
    if value > max(lst) or value < min(lst): # Only interpolation
        raise ValueError("This function does not allow for temperature extrapolation!")
    
    lst = np.array(lst)
    lower = list(lst[lst <= value]) # = sign to ensure non empty array
    higher = list(lst[lst >= value])

    low_closest = max(lower)
    high_closest = min(higher)
    return low_closest, high_closest
    
#*****************************************************************************************
#Function to compute thermochemical properties for 2nd order group contribution

def get_fit_value(all_models: dict, 
                            quantity: str,
                            temperature: int | float,
                            temperatures: list,
                            molecule) -> float:
    temp_low, temp_high = find_two_closest(temperature, temperatures)


    if abs(temp_high - temp_low) > 1e-12: # If input temp is element of temperatures: temp_low = temp_high
        weight_low = (temp_high - temperature)/(temp_high - temp_low) # for linear interpolation
    else:
        weight_low = 1
    weight_high = 1 - weight_low
    #ss = [getTypesFirstorder(Chem.MolFromSmiles(iupac_smiles(molecule)))]
    ss = [get_second_order_feature(iupac_smiles(molecule))]

    fit_low = all_models[(quantity, temp_low)].predict(ss)
    fit_high = all_models[(quantity, temp_high)].predict(ss)

    return weight_low * fit_low + weight_high * fit_high
    
#*****************************************************************************************
# function to obtain thermochemical properties from the training dataset (SI1.xlsx)
# SI1.xlsx must be in the same folder as this code

def get_scott_data(df,
                   molecule: str,
                   temperature: int | float,
                   quantity: str) -> float:
    
    dftemp = df[df['Molecule'] == iupac_smiles(molecule)]
    
    dftemp = dftemp[dftemp['Molecule'] == iupac_smiles(molecule)] # filter to molecule
    dftemp = dftemp[dftemp['T'] == temperature] # filter to temperature

    return dftemp[quantity].values # return the requested quantity

#*****************************************************************************************
 



